"""
*********************************************************************************
*                                                                               *
* http_request.py -- Methods to make single and multiple http requests          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import requests

from json.decoder import JSONDecodeError
from simplejson.errors import JSONDecodeError as SimpleJSONDecodeError

from black_widow.app.env import APP_VERSION, APP_NAME, APP_DEBUG
from black_widow.app.services import Log
from black_widow.app.helpers.validators import is_url


class HttpRequest:
    """
    HttpRequest Manager
    """

    DEFAULT_TIMEOUT = 30    # seconds

    class Type:
        """ Request Types """
        GET = 'get'
        POST = 'post'
        PUT = 'put'
        PATCH = 'patch'
        DELETE = 'delete'

        @staticmethod
        def all() -> tuple:
            """
            :return: all supported request methods
            """
            return HttpRequest.Type.GET, \
                HttpRequest.Type.POST, \
                HttpRequest.Type.PUT, \
                HttpRequest.Type.PATCH, \
                HttpRequest.Type.DELETE

    @staticmethod
    def request(
            url: str,
            request_type: str = Type.GET,
            data=None,
            json: dict or list = None,
            headers: dict = None,
            timeout: int = DEFAULT_TIMEOUT,
            cookies: str or dict = None
    ) -> requests.Response or None:
        """
        Make a request to chosen url
        :param url: The target url
        :param request_type: get|post|put|patch|delete
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`
        :param json: (optional) json data to send in the body of the :class:`Request`
        :param headers: The headers to send
        :param timeout: The request timeout
        :param cookies: The request cookies
        :return The response of request, or None (if the request fail)
        """
        if headers is None:
            headers = {}
        if type(cookies) is str:
            try:
                cookies = dict((k.strip(), v.strip()) for k, v in (c.split('=') for c in cookies.split(';')))
            except ValueError:
                # Wrong or empty cookies
                cookies = None
                pass

        req_headers = {
            'User-Agent': str(APP_NAME) + ' ' + str(APP_VERSION)
        }
        req_headers.update(headers)

        if data is None:
            data = {}

        request_type = request_type.lower()
        if not is_url(url):
            Log.error(str(url) + ' is not a valid url!')
            return None
        try:
            if request_type == HttpRequest.Type.GET:
                response = requests.get(url, data, headers=req_headers, timeout=timeout, cookies=cookies)
            elif request_type == HttpRequest.Type.POST:
                response = requests.post(url, data, json, headers=req_headers, timeout=timeout, cookies=cookies)
            elif request_type == HttpRequest.Type.PUT:
                response = requests.put(url, data, headers=req_headers, timeout=timeout, cookies=cookies)
            elif request_type == HttpRequest.Type.PATCH:
                response = requests.patch(url, data, headers=req_headers, timeout=timeout, cookies=cookies)
            elif request_type == HttpRequest.Type.DELETE:
                response = requests.delete(url, headers=req_headers, timeout=timeout, cookies=cookies)
            else:
                Log.error(str(request_type) + ' is not a valid request type!')
                return None
            if APP_DEBUG:
                HttpRequest.print_response(response)
            return response
        except (
                requests.exceptions.ConnectionError,
                requests.exceptions.TooManyRedirects,
                requests.exceptions.ReadTimeout
        ) as e:
            Log.error('Unable to complete request to ' + str(url))
            Log.error('Exception: ' + str(e))
        return None

    @staticmethod
    def multi_sequential_requests(
            urls: list,
            request_type: str = Type.GET,
            data=None,
            json: dict or list = None,
            headers: dict = None
    ) -> dict:
        """
        Make multiple sequential requests
        :param urls: The list of target urls
        :param request_type: get|post|put|patch|delete
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`
        :param json: (optional) json data to send in the body of the :class:`Request`
        :param headers: The headers to send
        :return A dictionary of responses like {'url_1': <response>, 'url_2': <response>, ...}
        """
        if APP_DEBUG:
            Log.info('CALLED: multi_request(' + str(urls) + ', ' + str(request_type) + ', ' + str(data) + ')')
        request_type = request_type.lower()
        response_dict = dict()
        for url in urls:
            response = HttpRequest.request(url, request_type, data, json, headers)
            if response is None:
                continue
            response_dict[url] = response
            if APP_DEBUG:
                try:
                    print(response.json())
                except (JSONDecodeError, SimpleJSONDecodeError):
                    print(response.text)
        return response_dict

    @staticmethod
    def is_image(response: requests.Response) -> bool:
        return HttpRequest.__check_header(response, 'content-type', 'image')

    @staticmethod
    def default_agent() -> str:
        """
        :return The black-widow agent
        """
        return str(APP_NAME) + ' ' + str(APP_VERSION)

    @staticmethod
    def print_response(response, limit=1000):
        """
        :param response: The response to print
        :param limit: The limit data length before truncate that
        """
        Log.info(str(response.url))
        Log.info('      |--- status_code: ' + str(response.status_code))
        Log.info('      |--- encoding: ' + str(response.encoding))
        Log.info('      |--- headers:')
        for key, value in response.headers.items():
            Log.info('      |       |--- ' + str(key) + ': ' + str(value))
        Log.info('      |')
        try:
            json_body = response.json()
            Log.info('      |-- data: ' + str(json_body))
        except ValueError:
            data = str(response.text)
            if len(data) > limit:
                data = '[truncated]' + data[0:limit]
            Log.info('      |-- data: ' + data)

    @staticmethod
    def __check_header(response: requests.Response, header: str, header_value: str) -> bool:
        if response.headers is None:
            return False
        for key, value in response.headers.items():
            key: str
            value: str
            if key.lower() != header.lower():
                continue
            return value.lower().split('/')[0] == header_value.lower()
        return False



"""
*********************************************************************************
*                                                                               *
* md5_crypto.py -- Md5 Encryption/Decryption manager                            *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import hashlib
import json

from black_widow.app.managers.request import HttpRequest
from black_widow.app.services import Log


class Md5Crypto:
    """
    Md5Crypto Manager
    """

    class Api:
        """
        Md5 APIs
        """

        # Method to get response data from 1st API
        @staticmethod
        def _api_1_result(json_dict):
            return json_dict.get('result')
        _api_1 = {'url': 'https://md5.pinasthika.com/api/decrypt?value=', 'get_result': _api_1_result}

        # Method to get response data from 2nd API
        @staticmethod
        def _api_2_result(json_dict):
            return json_dict[0].get('decrypted')
        _api_2 = {'url': 'https://www.md5.ovh/index.php?result=json&md5=', 'get_result': _api_2_result}

        @staticmethod
        def all():
            return Md5Crypto.Api._api_1, Md5Crypto.Api._api_2

    @staticmethod
    def encrypt(text: str):
        m = hashlib.md5()
        m.update(text.encode())
        return str(m.hexdigest())

    @staticmethod
    def decrypt(text: str):
        for api in Md5Crypto.Api.all():
            r = HttpRequest.request(api['url'] + text)
            if r is None:
                continue

            try:
                r_json = r.json()
            except json.decoder.JSONDecodeError:
                continue

            result = api['get_result'](r_json)
            if result is not None:
                return result
        Log.error('md5: unable to decrypt: ' + text)
        return None



#!/usr/bin/env python3

from setuptools import setup, find_namespace_packages

black_widow_src = "src/black_widow"

with open(black_widow_src + "/README.md", "r") as fh:
    long_description = fh.read()

with open(black_widow_src + "/requirements.txt", "r") as fh:
    install_requires = fh.readlines()

setup(
    name="black-widow",
    version="1.8.3",
    author="Fabrizio Fubelli",
    author_email="fabrizio@fubelli.org",
    description="Offensive penetration testing tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://black-widow.it",
    packages=find_namespace_packages(
        'src',
        include=(
            'black_widow',
            'black_widow.app',
            'black_widow.app.*',
            'black_widow.migrations',
            'black_widow.migrations.*',
            'black_widow.resources'
        ),
        exclude=('black_widow.docs', 'black_widow.docker', 'black_widow.lessons')
    ),
    package_dir={
        '': 'src'
    },
    include_package_data=True,
    exclude_package_data={
        'black_widow': ['docs', 'docker', 'resources']
    },
    package_data={
        '': [
            '*.html',
            '*.css',
            '*.js',
            '*.eot', '*.svg', '*.ttf', '*.woff', '*.woff2',
            '*.png', '*.jpg', '*.ico',
            # 'web.wsgi',
            'LICENSE.txt'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Natural Language :: English",
        "Topic :: Education :: Testing",
        'Topic :: Software Development :: Build Tools',
        "Topic :: System :: Clustering",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Networking",
        "Topic :: System :: Hardware :: Symmetric Multi-processing",
        "Topic :: Utilities"
    ],
    entry_points={
        'console_scripts': [
            'black-widow = black_widow:main',
        ],
        'gui_scripts': [
            'black-widow-gui = black_widow:main_gui',
        ]
    },
    python_requires='>=3.6',
    keywords='black-widow penetration testing offensive cyber security pentest sniffing',
    project_urls={
        'Documentation': 'https://github.com/offensive-hub/black-widow',
        'Source': 'https://github.com/offensive-hub/black-widow',
        'Tracker': 'https://github.com/offensive-hub/black-widow/issues',
    },
    install_requires=install_requires
)



"""
*********************************************************************************
*                                                                               *
* input_arguments.py -- Methods to parse the user input arguments.              *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import sys
import argparse

from black_widow import app


def create_parser():
    """
    Set the expected arguments
    :return: The created ArgumentParser
    """
    # --- Parser ---#
    parser = ArgumentParser(usage=sys.argv[0] + ' [Options]')

    # --- Options ---#
    options = parser.add_argument_group("Options")
    options.add_argument("-h", "--help", help="Show this help message and exit",
                         action="store_true")
    options.add_argument("-v", "--version", help="Show program's version number and exit",
                         action="store_true")
    # options.add_argument("-t", "--test", help="Execute a test",
    #                      action="store_true")
    options.add_argument("-g", "--gui", help="Run " + app.env.APP_NAME + " with GUI",
                         action="store_true")
    options.add_argument("--django", help="Run django manager (eg. `--django 'help'`)",
                         type=str, metavar='ARGS')

    # Packet Sniffing
    options_pcap = options.add_argument_group("Sniffing")
    options_pcap.add_argument("--pcap", help="Sniff Packages", action="store_true")
    options_pcap.add_argument("--pcap-src", help="The .pcap source file", type=argparse.FileType('r'),
                              metavar='FILE')
    options_pcap.add_argument("--pcap-dest", help="The .pcap destination file",
                              type=argparse.FileType('w'), metavar='FILE')
    options_pcap.add_argument("--pcap-int", help="Network interfaces (eg: eth0,wlan0)", type=str, metavar='INTERFACES')
    options_pcap.add_argument("--pcap-filters", help="https://wiki.wireshark.org/CaptureFilters", type=str,
                              metavar='FILTERS')
    options_pcap.add_argument("--pcap-limit", help="Max field lengths of each packet", type=int, metavar='INTEGER')
    options_pcap.add_argument("--pcap-count", help="Max packets to sniff", type=int, metavar='INTEGER')

    # Web Parsing
    crawl_types = '|'.join(app.managers.parser.HtmlParser.types())
    options_sql = options.add_argument_group("Web Parsing")
    options_sql.add_argument("--crawl", help="Crawl a website", action="store_true")
    options_sql.add_argument("--crawl-url", help="The url to crawl", type=str, metavar='URL')
    options_sql.add_argument("--crawl-type", help=crawl_types, type=str, metavar='TYPE')
    options_sql.add_argument("--crawl-depth", help="The crawl depth", type=int, metavar='INTEGER', default=0)

    # SQL Injection
    options_sql = options.add_argument_group("SQL Injection")
    options_sql.add_argument("--sql", help="Try injection in a website", action="store_true")
    options_sql.add_argument("--sql-url", help="The url where search for forms or to inject", type=str, metavar='URL')
    options_sql.add_argument("--sql-depth", help="Max crawling depth", type=int, metavar='INTEGER', default=0)
    options_sql.add_argument("--sql-forms", help="Parse the forms on page", action="store_true")

    return parser


# Creates the argument parser and return the parsed input arguments
def get_arguments():
    """
    Get the CLI arguments
    :return: dict The arguments
    """
    # --- Header image ---#
    def print_header():
        """
        Print the software header
        """
        header_ascii = app.helpers.storage.read_file(
            app.env.RES_PATH + '/' + str(app.env.APP_PROC) + '-ascii.txt'
        )
        header_ascii = header_ascii.replace('{version}', app.env.APP_VERSION)
        print('\n' + header_ascii + '\n')

    parser = create_parser()
    args = parser.parse_args()

    if args.version:
        print(app.env.APP_VERSION)
        sys.exit(0)

    print_header()

    # Check if at least one argument is set
    arg_set = False
    # noinspection PyProtectedMember
    for arg in args._get_kwargs():
        arg_set = arg[1] or arg_set
    if args.help or not arg_set:
        # Print help
        parser.print_help()
        print()  # print newline
        sys.exit(0)

    return args


def get_spaced_line(line, depth=0):
    """
    Prepend a line with certain space
    :param line: the line to edit
    :param depth: the number of spaces to insert
    :return: the prepended line
    """
    if type(line) == str:
        return (' ' * depth) + str(line)
    return line


class _CapitalisedHelpFormatter(argparse.HelpFormatter):

    # max_help_position
    def __init__(self, prog, indent_increment=2, max_help_position=36, width=None):
        argparse.HelpFormatter.__init__(self, prog, indent_increment, max_help_position, width)

    # CamelCase prefix
    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = 'Usage: '
        return super(_CapitalisedHelpFormatter, self).add_usage(usage, actions, groups, prefix)

    # Added depth
    def add_argument(self, action, depth=0):
        option_strings = []
        for option_string in action.option_strings:
            option_strings.append((' ' * depth) + option_string)
        action.option_strings = option_strings
        argparse.HelpFormatter.add_argument(self, action)

    # Added depth
    def add_arguments(self, actions, depth=0):
        for action in actions:
            self.add_argument(action, depth)


# Better subgroup managing (with depth)
class ArgumentParser(argparse.ArgumentParser):
    # noinspection PyDefaultArgument
    def __init__(self,
                 prog=None,
                 usage=None,
                 description=None,
                 epilog=None,
                 parents=[],
                 formatter_class=_CapitalisedHelpFormatter,
                 prefix_chars='-',
                 fromfile_prefix_chars=None,
                 argument_default=None,
                 conflict_handler='error',
                 add_help=False):
        argparse.ArgumentParser.__init__(self,
                                         prog=prog,
                                         usage=usage,
                                         description=description,
                                         epilog=epilog,
                                         # version=version,     # unexpected keyword (?!)
                                         parents=parents,
                                         formatter_class=formatter_class,
                                         prefix_chars=prefix_chars,
                                         fromfile_prefix_chars=fromfile_prefix_chars,
                                         argument_default=argument_default,
                                         conflict_handler=conflict_handler,
                                         add_help=add_help)

    # Overwrite Parent.format_help
    def format_help(self, depth=0):
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        # new recursive method to manage subgroups
        ArgumentParser.fill_formatter(self, formatter, depth)

        # epilog
        formatter.add_text(get_spaced_line(self.epilog, depth))

        return formatter.format_help()

    # noinspection PyProtectedMember
    @staticmethod
    def fill_formatter(argument, formatter, depth=0):
        """
        Manage correctly the subgroups depth
        :param argument: The argument to format
        :param formatter: The formatter to fill
        :param depth: The initial depth of the printed params
        """
        actions = argument._action_groups
        if hasattr(argument, '_group_actions'):
            actions += argument._group_actions

        # print(type(argument))

        if type(argument) == ArgumentParser:
            subsections = True
        else:
            subsections = False
            formatter.start_section(get_spaced_line(argument.title, depth))

        for action_group in actions:
            if subsections:
                formatter.start_section(get_spaced_line(action_group.title, depth))
                formatter.add_text(get_spaced_line(action_group.description, depth))
                formatter.add_arguments(action_group._group_actions, depth=depth)
                for child_action_group in action_group._action_groups:
                    if type(child_action_group) == argparse._ArgumentGroup:
                        ArgumentParser.fill_formatter(child_action_group, formatter, depth)
            else:
                formatter.add_argument(action_group, depth=depth)

            if subsections:
                formatter.end_section()

        if not subsections:
            formatter.end_section()



"""
*********************************************************************************
*                                                                               *
* sqlmaptask.py -- sqlmap task.                                                 *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import requests

from black_widow.app.managers.request import HttpRequest
from black_widow.app.services import Log, JsonSerializer


class SqlmapTask:
    """
    The SqlmapTask class
    """

    __TASKS = dict()

    def __init__(self, task_id: str, api_url: str, target_url: str):
        self.id = task_id
        self.api_url = api_url
        self.target_url = target_url
        if SqlmapTask.__TASKS.get(api_url) is None:
            SqlmapTask.__TASKS[api_url] = dict()
        SqlmapTask.__TASKS[api_url][task_id] = self

    """ Public static methods """

    ##########################
    # Sqlmap Admin functions #
    ##########################

    @staticmethod
    def task_list(api_url: str):
        """
        Pull task list
        :param api_url: The base_url of sqlmap-api (eg. "http://127.0.0.1:8775")
        :rtype: dict
        """
        r_data = SqlmapTask._request(api_url + '/admin/list')
        task_dict = SqlmapTask.__TASKS.get(api_url)
        active_task_keys = r_data['tasks'].keys()
        if task_dict is None:
            return dict()
        for task_id, task in task_dict.items():
            if task_id not in active_task_keys:
                SqlmapTask.__TASKS.pop(task_id)
        return SqlmapTask.__TASKS

    @staticmethod
    def task_flush(api_url: str):
        """
        Flush task spool (delete all tasks)
        :param api_url: The base_url of sqlmap-api (eg. "http://127.0.0.1:8775")
        """
        SqlmapTask._request(api_url + '/admin/flush')

    #############################
    # Task management functions #
    #############################

    @staticmethod
    def task_new(api_url: str, target_url: str):
        """
        Create a new task
        :param api_url: The base_url of sqlmap-api (eg. "http://127.0.0.1:8775")
        :param target_url: The target url
        :rtype: SqlmapTask
        """
        r_data = SqlmapTask._request(api_url + '/task/new')
        return SqlmapTask(r_data['taskid'], api_url, target_url)

    ###############
    # Delete task #
    ###############

    def task_delete(self):
        """
        Delete this existing task
        """
        return self.__task_request('delete')

    ##################################
    # Sqlmap core interact functions #
    ##################################

    def option_list(self) -> dict:
        """
        List options for this task
        """
        return self.__option_request('list')

    def option_get(self, options: list) -> dict:
        """
        Get value of option(s) for this task
        :param options: The options to get (eg. [ "cookie", "headers", "referer", ... ])
        """
        return self.__option_request('get', HttpRequest.Type.POST, options)

    def option_set(self, options: dict) -> dict:
        """
        Get value of option(s) for this task
        :param options: The options to set (eg. { "referer": "https://example.com" ])
        """
        return self.__option_request('set', HttpRequest.Type.POST, options)

    ################
    # Handle scans #
    ################

    def scan_start(self):
        """
        Launch the scan
        """
        return self.__scan_request('start', HttpRequest.Type.POST, {})

    def scan_stop(self):
        """
        Stop the scan
        """
        return self.__scan_request('stop')

    def scan_kill(self) -> dict:
        """
        Kill the scan
        """
        return self.__scan_request('kill')

    def scan_status(self) -> dict:
        """
        Returns status of the scan
        """
        return self.__scan_request('status')

    def scan_data(self) -> dict:
        """
        Retrieve the data of the scan
        """
        return self.__scan_request('data')

    def scan_log(self) -> dict:
        """
        Retrieve the log messages of the scan
        """
        return self.__scan_request('log')

    ###################
    # Utils functions #
    ###################

    @staticmethod
    def _request(url: str, request_type: str = HttpRequest.Type.GET, json: dict or list = None) -> dict:
        """
        Send a request to sqlmap-api server and then load the data json as dict
        :param url: The sqlmap-api url (eg. "http://127.0.0.1:8775/task/new")
        :param request_type: get|post|put|patch|delete
        :param json: The json to send
        :rtype: dict
        """
        response = HttpRequest.request(url, request_type, json=json)
        r_data = JsonSerializer.load_json(response.text)
        Log.info('Response data of ' + url + ': ' + str(r_data))
        success = r_data.get('success')
        if success is None:
            # The response has not the status management
            return r_data
        if not success:
            Log.error('Response data of ' + url + ' has { success: False }')
            raise requests.RequestException('Request to ' + url + ' failed')
        return r_data

    """ Private methods """

    def __request(self, path: str, request_type: str = HttpRequest.Type.GET, json: dict or list = None) -> dict:
        """
        :param path: The path for request (eg. "/task/<id>/start")
        :param request_type: get|post|put|patch|delete
        :param json: The json to send
        :rtype: dict
        """
        url = self.api_url + path
        return SqlmapTask._request(url, request_type, json)

    def __option_request(
            self,
            action: str,
            request_type: str = HttpRequest.Type.GET,
            json: dict or list = None
    ) -> dict:
        """
        :param action: The action of option request (eg. "list")
        :param request_type: get|post|put|patch|delete
        :param json: The json to send
        :rtype: dict
        """
        return self.__request('/option/' + self.id + '/' + action, request_type, json)

    def __task_request(self, action: str, request_type: str = HttpRequest.Type.GET, json: dict or list = None) -> dict:
        """
        :param action: The action of task request (eg. "delete")
        :param request_type: get|post|put|patch|delete
        :param json: The json to send
        :rtype: dict
        """
        return self.__request('/task/' + self.id + '/' + action, request_type, json)

    def __scan_request(self, action: str, request_type: str = HttpRequest.Type.GET, json: dict or list = None) -> dict:
        """
        :param action: The action of scan request (eg. "kill")
        :param request_type: get|post|put|patch|delete
        :param json: The json to send
        :rtype: dict
        """
        return self.__request('/scan/' + self.id + '/' + action, request_type, json)



"""
*********************************************************************************
*                                                                               *
* manage.py -- Django manager.                                                  *
*                                                                               *
* Methods to use the Django manager to test the web server                      *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os
import sys
import webbrowser

from gunicorn.app.wsgiapp import run as gunicorn_run
from django.core import management

from black_widow.app.env import EXEC_PATH, APP_NAME, APP_WEB_HOST, APP_WEB_PORT
from black_widow.app.services import MultiTask, Log
from black_widow.app.helpers.network import get_ip_address
from black_widow.app.helpers.storage import delete
from black_widow.app.helpers.util import whoami, set_owner_process
from black_widow.app.gui.web.wsgi import WEB_PACKAGE


def django_gui():
    sys.path.insert(0, os.path.dirname(__file__))
    bind_host = _get_bind_socket()
    Log.info("Starting " + str(APP_NAME) + ' GUI')
    sys.argv = [sys.argv[0], 'web.wsgi', '-b', bind_host]
    django_cmd(['migrate'])
    _launch_browser(bind_host)
    gunicorn_run()


def django_cmd(args: list):
    if 'runserver' not in args:
        # cd to "web" directory
        os.chdir(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'web'))
    elif len(args) <= 1:
        bind_host = _get_bind_socket()
        args.append(bind_host)
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", WEB_PACKAGE + ".settings")
    management.execute_from_command_line([EXEC_PATH + ' --django', ] + args)


def _launch_browser(bind_host: str):
    user = whoami()

    def browser_target():
        """
        The function that launch the browser
        """
        set_owner_process(user)
        Log.info('Launching browser with User: ' + str(whoami(False)))
        webbrowser.open('http://' + bind_host)
        Log.success('Web browser opened')

    pidfile = MultiTask.multiprocess(browser_target, asynchronous=True, cpu=1)
    delete(pidfile)     # The pidfile is not required


def _get_bind_socket():
    host = APP_WEB_HOST
    if host is None:
        host = get_ip_address()
    if host is None:
        host = '0.0.0.0'
    return str(host) + ':' + str(APP_WEB_PORT)



"""
*********************************************************************************
*                                                                               *
* sniffing_view.py -- Django Sniffing views.                                    *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.http import JsonResponse
from django.shortcuts import render, redirect

from black_widow.app.gui.web.black_widow.models import SniffingJobModel
from black_widow.app.helpers import network, util
from black_widow.app.managers.sniffer import PcapSniffer

from .abstract_sniffing_view import AbstractSniffingView


class Sniffing:
    """
    Sniffing Container View
    """
    class SettingsView(AbstractSniffingView):
        """
        Sniffing Settings View
        """
        name = 'sniffing'
        template_name = 'sniffing/settings.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            if not PcapSniffer.is_executable():
                return render(request, self.error_templates.get('root_required'))
            view_params = {
                'interfaces': network.get_interfaces(),
                'jobs': SniffingJobModel.all()
            }
            return render(request, self.template_name, view_params)

        def post(self, request):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponseRedirect
            """
            if not PcapSniffer.is_executable():
                return JsonResponse({
                    'message': 'You are not #root'
                }, status=401)
            request_params: dict = request.POST.dict()
            if request_params.get('interfaces') is not None:
                request_params['interfaces'] = request.POST.getlist('interfaces')
            pcap_file = request.FILES.get('pcap')
            if pcap_file is not None:
                request_params['pcap'] = self.upload_file(pcap_file)
            else:
                request_params['pcap'] = None

            job = self._new_job(
                request_params.get('filters'),
                request_params.get('pcap'),
                request_params.get('interfaces'),
            )

            return redirect('/sniffing/capture?id=' + str(job.id))

    class CaptureView(AbstractSniffingView):
        """
        Sniffing Capture View
        """
        name = 'sniffing'
        template_name = 'sniffing/capture.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            if not PcapSniffer.is_executable():
                return render(request, self.error_templates.get('root_required'))

            return self._get_job(request, redirect_url='/sniffing')

        def post(self, request) -> JsonResponse:
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.JsonResponse
            """
            return self._post_job(request)



"""
*********************************************************************************
*                                                                               *
* sql_injection.py -- Methods to try injection in forms.                        *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from pprint import pprint
from time import sleep
from urllib.parse import urlparse

from black_widow.app.managers.injection.sql_injection_util.sqlmaptask import SqlmapTask
from black_widow.app.managers.parser import HtmlParser
from black_widow.app.services import Log

from .sql_injection_util import SqlmapClient


class SqlInjection:
    """
    SqlInjection Manager
    """

    @staticmethod
    def inject(
            forms: bool,
            url: str = None,
            max_depth: int = None,
            user_callback: callable = None
    ) -> dict:
        """
        Try SQL injection
        :param forms: True if you want to search forms in provided url
        :param url: The url to inject, or the url to visit in search for forms
        :param max_depth: The max crawling depth (if forms is False, it is skipped)
        :param user_callback: The callback method to call (or None) (@see SqlInjection._user_callback_example)
        :return: A dictionary of SqlmapTask
        """
        if not forms:
            # Inject directly the provided url
            task = SqlInjection.__inject_url(url)
            tasks = {
                task.id: task
            }
        else:
            tasks = SqlInjection.__inject_forms(url, max_depth)
        # Search forms inside the provided url
        if user_callback is None:
            SqlInjection.__listen_tasks(tasks)
        else:
            return user_callback(tasks)

    # noinspection PyUnusedLocal
    @staticmethod
    def _user_callback_example(tasks: dict):
        """
        This is an user_callback example method.
        It manages the tasks dictionary
        :param tasks: A dictionary of SqlmapTask
        :return: None
        """
        # Manage tasks
        return None

    @staticmethod
    def __inject_url(url: str) -> SqlmapTask:
        return SqlmapClient.try_inject_url(url)

    @staticmethod
    def __inject_forms(url, max_depth) -> dict:
        """
        Search a form in the page returned by url.
        If it doesn't find a form, or the injection can't be done, it visit the website in search for other forms
        :param url: str The url to visit
        :param max_depth: int The max depth during the visit
        :return A dictionary of SQL injection tasks
        """

        base_url = urlparse(url).netloc
        parsed_forms = dict()
        tasks = dict()

        def deep_inject_form(href, depth=0):
            # Check the domain
            if href in parsed_forms or \
                    urlparse(href).netloc != base_url or \
                    (max_depth is not None and depth > max_depth):
                return ''

            # Visit the current href
            parsed_relevant, request_cookies = HtmlParser.relevant_parse(href)

            # Find forms in page
            parsed_forms[href] = HtmlParser.find_forms(parsed_relevant, href)

            # Execute Sqlmap task
            task = SqlmapClient.try_inject_form(href, parsed_forms, request_cookies)
            tasks[task.id] = task
            Log.success('SQL injection of "' + href + '" started!')

            # Find adjacent links
            links = HtmlParser.find_links(parsed_relevant)

            # Visit adjacent links
            for link in links:
                # print('link: '+link)
                child_request_cookies = deep_inject_form(link, depth+1)
                if len(child_request_cookies) > len(request_cookies):
                    request_cookies = child_request_cookies

        Log.success('SQL injection started!')
        deep_inject_form(url)
        Log.success('Website crawled! Found '+str(len(parsed_forms))+' forms')
        return tasks

    @staticmethod
    def __listen_tasks(tasks: dict):
        task_statuses = dict()
        while True:
            sleep(3)
            for task_id, task in tasks.items():
                task_status = str(task.scan_data()) + str(task.scan_log())
                if task_statuses.get(task_id) == task_status:
                    continue
                task_statuses[task_id] = task_status
                SqlInjection.__print_task(task)

    @staticmethod
    def __print_task(task: SqlmapTask):
        scan_log = task.scan_log()
        scan_data = task.scan_data()
        print('')
        print('----------------------------------------------------------')
        print('[SQL injection] Scan Url: ' + task.target_url)
        print('----------------------------------------------------------')
        print('[SQL injection] Scan Log:')
        pprint(scan_log)
        print('----------------------------------------------------------')
        print('[SQL injection] Scan Data:')
        print('----------------------------------------------------------')
        pprint(scan_data)
        print('')
        print('')



"""
*********************************************************************************
*                                                                               *
* base64_crypto.py -- Base64 Encoding/Decoding manager                          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from base64 import b64decode, b64encode


class Base64Crypto:
    """
    Base64Crypto Manager
    """

    @staticmethod
    def encrypt(text: str) -> str:
        return b64encode(bytes(text, encoding='utf-8')).decode('utf-8')

    @staticmethod
    def decrypt(text: str):
        return b64decode(text).decode('utf-8')



"""
*********************************************************************************
*                                                                               *
* serializer.py -- The serializer classes for global usage.                     *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import json
import os
import pickle

from black_widow.app.services import Log


class PickleSerializer:
    """ The black-widow PickleSerializer """

    @staticmethod
    def get_object(file: str):
        """
        :param file: A file that contains a dumped object
        :return:
        """
        if not os.path.isfile(file):
            return None
        f = open(file, 'rb')
        obj = pickle.load(f)
        f.close()
        return obj

    @staticmethod
    def set_object(obj, file: str):
        """
        :param obj: The object to dump in file
        :param file: The file where dumps the object
        """
        f = open(file, 'wb')
        pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)
        f.close()

    @staticmethod
    def add_item_to_dict(key, value, file: str):
        """
        :param key: The dictionary key
        :param value: The dictionary value
        :param file: The file where dictionary is dumped
        """
        dictionary = PickleSerializer.get_object(file)
        if type(dictionary) != dict:
            dictionary = dict()
        dictionary[key] = value
        PickleSerializer.set_object(dictionary, file)


class JsonSerializer:
    """ The black-widow JsonSerializer """

    @staticmethod
    def get_dictionary(file: str) -> dict:
        """
        :param file: A file that contains a json
        :return: A dictionary
        """
        if not os.path.isfile(file):
            Log.error(file + ' is not a file')
            return dict()
        return JsonSerializer.load_json(file)

    @staticmethod
    def set_dictionary(dictionary: dict, file: str):
        """
        :param dictionary: The dictionary to dump in file
        :param file: The file where dumps the object
        """
        JsonSerializer.dump_json(dictionary, file)

    @staticmethod
    def add_item_to_dict(key, value, file: str):
        """
        :param key: The dictionary key or None
        :param value: The dictionary value
        :param file: The file where dictionary is dumped
        """
        dictionary = JsonSerializer.get_dictionary(file)
        if key is None:
            key = len(dictionary)
        dictionary[key] = value
        JsonSerializer.set_dictionary(dictionary, file)

    @staticmethod
    def dump_json(obj, file: str):
        """
        Write the input object into the input file
        :param file: The output json file
        :type obj: dict or list
        :return: The dumped json of object
        """
        with open(file, 'w') as outfile:
            json.dump(obj, outfile, indent=2)

    @staticmethod
    def dumps_json(obj) -> str:
        """
        Convert the input object into a json string
        :type obj: dict or list
        :return: The dumped json of object
        """
        try:
            return json.dumps(obj)
        except json.decoder.JSONDecodeError:
            return ""

    @staticmethod
    def load_json(file: str) -> dict:
        """
        :param file: The file to read
        :return: A dictionary
        """
        try:
            with open(file, 'r') as infile:
                return json.load(infile)
        except json.decoder.JSONDecodeError as e:
            Log.error(str(e))
            return dict()



"""
*********************************************************************************
*                                                                               *
* abstract_job_model.py -- A sniffing job info.                                 *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os
import signal

from datetime import datetime
from time import sleep

from django.db import models
from django.utils.timezone import now

from black_widow.app.gui.web.black_widow.models.abstract_model import AbstractModel
from black_widow.app.helpers import storage
from black_widow.app.helpers.util import pid_exists, sort_dict, reverse_dict
from black_widow.app.services import MultiTask, Log, JsonSerializer


class AbstractJobModel(AbstractModel):
    """
    Django Abstract Job Model
    """
    json_file: str = models.CharField(max_length=250, null=False)
    status: int = models.PositiveIntegerField(null=False, default=signal.SIGCONT)
    pid: int = models.PositiveIntegerField(null=False)
    _pid_file: str = models.CharField(max_length=250, null=False)
    created_at: datetime = models.DateTimeField(default=now, editable=False)
    json_sort_value: str = None

    class Meta:
        abstract = True

    @staticmethod
    def _all(model_class) -> models.query.QuerySet:
        """
        :param model_class: A class that implements AbstractJobModel
        :return: All existent jobs of model_class
        """
        jobs = model_class.objects.all().order_by('-id')
        for job in jobs:
            job: AbstractJobModel
            job.self_check()
        return jobs

    def self_check(self):
        if self.status not in (signal.SIGKILL, signal.SIGABRT) and not pid_exists(self.pid):
            self.status = signal.SIGKILL
            self.save()

    @property
    def pid_file(self) -> str:
        return self._pid_file

    @pid_file.setter
    def pid_file(self, value: str):
        self._pid_file = value
        pids = MultiTask.get_pids_from_file(self.pid_file)
        if len(pids) != 1:
            raise ChildProcessError("Unable to find process PID of this Sniffing Job")
        self.pid = int(pids[0])

    @property
    def status_name(self) -> str:
        return signal.Signals(self.status).name

    def kill(self, sig: int):
        """
        Send a signal to process which is running this job
        :param sig: The signal as integer (eg. 9 for SIGKILL)
        """
        if self.status == sig:
            return
        Log.info("Sending signal " + str(sig) + " to job #" + str(self.id) + ' (' + str(self.pid) + ')')
        self.status = sig
        try:
            os.kill(self.pid, sig)
        except ProcessLookupError:
            # Process does not exists
            pass
        self.save()
        Log.success("Signal " + str(sig) + " sent to job #" + str(self.id) + ' (' + str(self.pid) + ')')

    def delete(self, using=None, keep_parents=False):
        self.kill(signal.SIGKILL)
        if not storage.delete(self.json_file):
            return False
        if not storage.delete(self.pid_file):
            return False
        return super(AbstractJobModel, self).delete(using, keep_parents)

    @property
    def json_dict(self) -> dict:
        attempts = 0
        json_dict = JsonSerializer.get_dictionary(self.json_file)
        while len(json_dict) == 0:
            sleep(0.2)
            # Prevents parallel access errors to file
            json_dict = JsonSerializer.get_dictionary(self.json_file)
            attempts += 1
            if attempts >= 5:
                break
        if self.json_sort_value is None:
            return reverse_dict(json_dict)
        return sort_dict(dict(sorted(
            json_dict.items(),
            key=lambda e: int(e[1][self.json_sort_value]),
            reverse=True
        )))

    @property
    def total(self) -> int:
        json_sort_value = self.json_sort_value
        self.json_sort_value = None
        total = len(self.json_dict)
        self.json_sort_value = json_sort_value
        return total
